/*
 *
 * Copyright (c) 2003, Frederick Ros (sl33p3r@free.fr)
 *                                            
 * eu_msg.c  - Message Procol Subsystem. Writen based on Analogic Devices
 *             supplied code.
 *
 * This file is part of the eagle-usb driver package.
 *
 * eagle-usb driver package is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * "eagle-usb driver package" is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: eu_msg.c,v 1.9 2004/11/01 10:33:26 sleeper Exp $
 */

#include "Adiutil.h"
#include "macros.h"
#include "eu_msg.h"
#include "eu_utils.h"
#include "debug.h"




/* -------------------------- Private Macros/Variables -------------------------- */

#define CDC_REQ_SEND_ENCAPSULATED_CMD 0x00
#define CMD_SET_BLOCK                 0x01
#define DSP_MP_TX_START               0x3FCE

#define SEARCHING   0x0
#define PASSED      0x1
#define FAILED      0x2

#define NUM_CMVS                20

#define PREAMBLE                0x535C
#define RECEIVER_MODEM          0x0000
#define SENDER_HOST             0x0010
#define SUBTYPE_MEM_REQWRITE    0x0100
#define SUBTYPE_MEM_REQREAD     0x0000
#define TYPE_MEMACCESS          0x1000

#define WRITE_TO_MODEM (RECEIVER_MODEM + SENDER_HOST + SUBTYPE_MEM_REQWRITE + TYPE_MEMACCESS)
#define READ_FRM_MODEM (RECEIVER_MODEM + SENDER_HOST + SUBTYPE_MEM_REQREAD  + TYPE_MEMACCESS)

unsigned short e4_mp_msg_encode 
(
	unsigned char ucRWFlag,
	uint16_t usGroup,
	uint16_t usAddress,
	uint16_t usIndex,
	uint32_t uiData,
	uint16_t usPayloadSize,
	uint16_t *mp_raw 
);


/* ---------------------------- Private Declarations ---------------------------- */

static void eu_write_dsp_msg (
                              eu_instance_t *ins,
                              uint16_t       addr,
                              uint16_t      *data
                             );

/* ----------------------------- Exported Functions ----------------------------- */

/**
 * eu_decode_msg - Decode messages coming from modem
 *
 * @msg     - storage for decoded message
 * @raw_msg - Raw message as sent by the modem
 *
 * returns a boolean
 *
 */
uint32_t eu_decode_msg ( eu_msg_t *msg, uint16_t raw_msg[8] ) 
{
    int       retcode = TRUE;
    uint16_t *pword   = NULL;
    uint16_t  tmp     = 0;
    
    pword = &raw_msg[0];
    
    /*
     * Get preamble first
     */
    msg->preamble = cpu_to_le16 (*pword);

    /*
     * And check it
     */
    if ( msg->preamble != MSG_PREAMBLE ) 
    {
        retcode = FALSE;
        eu_err ("eu_decode_msg: Invalid preamble in message\n");
        goto byebye;
    }

    
    /*
     * Get Function type
     */
    pword ++;
    
    tmp = cpu_to_le16 (*pword);

    msg->type = ( tmp & MP_FUNCTION_TYPE_MASK ) >> MP_FUNCTION_TYPE_LSB;

    /*
     * And check it ...
     */
    switch ( msg->type ) 
    {
        case MP_FUNCTION_TYPE_MEMACCESS:
        case MP_FUNCTION_TYPE_FLASHACC:
        case MP_FUNCTION_TYPE_ADSLDIRECTIVE:
        case MP_FUNCTION_TYPE_MSGDECERR:
        case MP_FUNCTION_TYPE_MSGACCERR:
        case MP_FUNCTION_TYPE_FLASHACCERR:
            /*
             * OK : legal types
             */
            break;
            
        default:
            /*
             * Illegal function type
             */
            retcode = FALSE;
            eu_err ("eu_decode_msg: Illegal function type: 0x%x\n",msg->type);
            goto byebye;
            break;
    }

    /*
     * Get function subtype
     */
    
    tmp = cpu_to_le16 ( *pword );

    msg->subtype = ( tmp & MP_FUNCTION_SUBTYPE_MASK ) >> MP_FUNCTION_SUBTYPE_LSB;

    /*
     * And check it
     */
    switch ( msg->type )
    {
        case MP_FUNCTION_TYPE_MSGACCERR:
            eu_warn ("eu_decode_msg: Received a message 0x%x (subtype 0x%x)",
                     msg->type,msg->subtype);
            break;
            
        case MP_FUNCTION_TYPE_MEMACCESS:
            if ( ( msg->subtype != SUBTYPE_MEMACCESS_REPLYREAD ) &&
                 ( msg->subtype != SUBTYPE_MEMACCESS_REPLYWRITE ) ) 
            {
                retcode = FALSE;
                eu_err ("eu_decode_msg: Illegal function subtype: 0x%x (type=MEMACCESS)\n",msg->subtype);
                goto byebye;
            }
            break;
            
        case MP_FUNCTION_TYPE_FLASHACC:
            switch ( msg->subtype ) 
            {
                case SUBTYPE_FLASHACC_ENKERNELREBOOTREPLY:
			    case SUBTYPE_FLASHACC_ENMODEMREBOOTREPLY:
			    case SUBTYPE_FLASHACC_DECOMPMODEMAPPLREPLY:
			    case SUBTYPE_FLASHACC_STREAMEDDATAWRREPLY:
			    case SUBTYPE_FLASHACC_STREAMEDDATAWRBLOCKREPLY:
			    case SUBTYPE_FLASHACC_READREPLY:
                    /*
                     * OK : legal subtypes
                     */
                    break;
                    
			    default:
                    retcode = FALSE;
                    eu_err ("eu_decode_msg: Illegal function subtype: 0x%x (type=FLASHACC)\n",msg->subtype);
                    goto byebye;
                    break;
            }
            break;

        case MP_FUNCTION_TYPE_FLASHACCERR:
            if ( ( msg->subtype != SUBTYPE_FLASHACCERR_DECOMPRESSERR ) &&
                 ( msg->subtype != SUBTYPE_FLASHACCERR_READCOMPERR ) ) 
            {
                retcode = FALSE;
                eu_err ("eu_decode_msg: Illegal function subtype: 0x%x (type=FLASHACCERR)\n",msg->subtype);
                goto byebye;
            }
            break;

        case MP_FUNCTION_TYPE_ADSLDIRECTIVE:
            if ( ( msg->subtype == SUBTYPE_ADSLDIRECTIVE_KERNELREADY ) ||
                 ( msg->subtype == SUBTYPE_ADSLDIRECTIVE_MODEMREADY ) ||
                 ( msg->subtype == SUBTYPE_ADSLDIRECTIVE_MODEMCRCERROR ) )
            {
                /*
                 * OK : legal subtypes
                 */                
                break;
            }
            
            retcode = FALSE;
            eu_err ("eu_decode_msg: Illegal function subtype: 0x%x (type=ADSLDIRECTIVE)\n",msg->subtype);
            goto byebye;
            break;
            
        default:
            retcode = FALSE;
            eu_err ("eu_decode_msg: Illegal function type: 0x%x\n",msg->type);
            goto byebye;
            break;
    }
    
    /*
     * Get Function Sender ID
     */
    tmp = cpu_to_le16 ( *pword );

    msg->sender_id = ( tmp & MP_FUNCTION_SENDERID_MASK ) >> MP_FUNCTION_SENDERID_LSB;

    if ( msg->sender_id != SUBTYPE_ID_ADSL ) 
    {
        retcode = FALSE;
        eu_err ("eu_decode_msg: Illegal sender id: 0x%x\n",msg->sender_id);
        goto byebye;
    }
    
    /*
     * Get receiver ID
     */
    tmp = cpu_to_le16 ( *pword );

    msg->receiver_id = ( tmp & MP_FUNCTION_RECEIVER_MASK ) >> MP_FUNCTION_RECEIVER_LSB;

    if ( msg->receiver_id != SUBTYPE_ID_ME ) 
    {
        retcode = FALSE;
        eu_err ("eu_decode_msg: Illegal receiver id: 0x%x\n",msg->receiver_id);
        goto byebye;
    }

    if ( msg->type == MP_FUNCTION_TYPE_ADSLDIRECTIVE ) 
    {
        msg->index = 0;
        msg->offset_addr = 0;
        msg->data = 0xDEADBEEF;
        retcode = TRUE;
        goto byebye;
    }

    /*
     * Get index
     */
    pword ++;

    msg->index = cpu_to_le16 ( *pword );

    /*
     * Get Symbolic address
     */
    pword ++;
    
    tmp = cpu_to_le16 ( *pword ); /* Addr High */
    msg->saddr[0] = (unsigned char) ( (unsigned short) ( tmp &0xFF00) >> 8);
    msg->saddr[1] = (unsigned char) ( (unsigned short) ( tmp &0x00FF));
    
    pword ++;
    tmp = cpu_to_le16 ( *pword ); /* Addr Low */
    msg->saddr[2] = (unsigned char) ( (unsigned short) ( tmp &0xFF00) >> 8);
    msg->saddr[3] = (unsigned char) ( (unsigned short) ( tmp &0x00FF));

    /*
     * Get Offset
     */
    pword++;
    msg->offset_addr = *pword;

    /*
     * And data
     */
    pword ++;
    tmp = cpu_to_le16 ( *pword ); /* Addr High */

    pword++;                    /* Addr Low */
    msg->data = tmp * 0x10000 + cpu_to_le16 ( *pword );
    
    
  byebye:
    return (retcode);
    
}


/**
 * eu_send_msg  -  Send given message ( message subsystem must have been initialized)
 *
 */
unsigned int eu_send_msg ( eu_instance_t *ins, uint16_t *packet ) 
{
    unsigned int ret_code = true;
    eu_enters (DBG_MSG);
    
    
    /*
     * if no message pending, its safe to send another one
     */
    /* FFD_DEBUG if ( ins->AdiModemSm.OutboundPending == FALSE )*/
    
    if ( ins->e4sm.mp_tx_msg_pending == false)
    {
        /*
         * Set the flag because the send is pending
         */
        /* FFD_DEBUG ins->AdiModemSm.OutboundPending = TRUE;*/
	ins->e4sm.mp_tx_msg_pending = true;
    
        eu_write_dsp_msg ( ins, DSP_MP_TX_START, (uint16_t *) packet );
    
    }
    else
    {
        ret_code = false;
        eu_dbg (DBG_MSG,"TxPending = TRUE, cannot send new message!\n");
        
    }
    eu_leaves (DBG_MSG);
    return ret_code;
}

/**
 * eu_msg_initialize - Initialize Message subsystem. Pre-encodes all necessary
 *                     messages, and store them in a global structure.
 *
 */

int e4_msg_initialize ( eu_instance_t *ins, const eu_options_t opt )
{
    EAGLE_IV *pSm = &(ins->e4sm);
    
    int counter = 0;
    
    eu_enters (DBG_MSG);
    
    pSm->LineType = opt[CFG_LINE].value;
    
	
	if (e4_mp_msg_encode('r',cmd_Group2, 0  ,0 , 0x0,1,&pSm->mp_statcmv[0])!= MP_ENCODE_OK)
	{
		return false; 
	}
        
    	counter = 0;

	if (e4_mp_msg_encode('r',cmd_Group6, 0  ,0 , 0x0,4,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK)
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 4;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_UR;
		counter++;
	}
	
	if (e4_mp_msg_encode('r',cmd_Group6, 1  ,0 , 0x0,4,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 4;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_DR;
		counter++;
		
	}

	if (e4_mp_msg_encode('r',cmd_Group3, 55  ,0 , 0x0,3,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 3;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_FV;
		counter++;
		
	}
	
	if (e4_mp_msg_encode('r',cmd_Group2, 1  ,0 , 0x0,1,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 1;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_AS;
		counter++;
		
	}
	
	if (e4_mp_msg_encode('r',cmd_Group3, 64 ,0 , 0x0,4,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 4;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_CV1;
		counter++;
	}
	
	if (e4_mp_msg_encode('r',cmd_Group3, 65 ,0 , 0x0,4,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 4;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_CPV1;
		counter++;
	}

	if (e4_mp_msg_encode('r',cmd_Group3, 68  ,1 , 0x0,1,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 1;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_NA;
		counter++;
		
	}
	
	if (e4_mp_msg_encode('r',cmd_Group3, 69  ,1 , 0x0,1,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 1;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_FA;
		counter++;
		
	}

	pSm->usNumOnetimeOperationalCmvs = counter;

	if (e4_mp_msg_encode('r',cmd_Group2, 0  ,0 , 0x0,1,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK)
	{
		return false; 
	}
	else
	{
	
		pSm->opCMVS[counter].ucPayloadSize = 1;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_LS;
		counter++;
		
	}
	
	if (e4_mp_msg_encode('r',cmd_Group7, 0  ,0 , 0x0,1,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 1;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_NFLG;
		counter++;
		
	}
	
	if (e4_mp_msg_encode('r',cmd_Group3, 68  ,3 , 0x0,1,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 1;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_NS;
		counter++;
	}

	
	if (e4_mp_msg_encode('r',cmd_Group3, 69  ,3 , 0x0,1,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 1;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_FS;
		counter++;
	}

	if (e4_mp_msg_encode('r',cmd_Group7, 11  ,0 , 0x0,4,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 4;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_NH;
		counter++;
	}
	
	if (e4_mp_msg_encode('r',cmd_Group7, 2 ,0 , 0x0,2,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 2;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_NC;
		counter++;
	}

	if (e4_mp_msg_encode('r',cmd_Group7, 24 ,0 , 0x0,2,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 2;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_FC;
		counter++;
	}

	if (e4_mp_msg_encode('r',cmd_Group7, 6 ,0 , 0x0,1,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 1;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_NF;
		counter++;
	}


	if (e4_mp_msg_encode('r',cmd_Group7, 32 ,0 , 0x0,1,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 1;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_FF;
		counter++;
	}


	if (e4_mp_msg_encode('r',cmd_Group7, 20 ,0 , 0x0,1,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 1;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_DLT;
		counter++;
	}


	if (e4_mp_msg_encode('r',cmd_Group7, 21 ,0 , 0x0,1,&pSm->opCMVS[counter].usMP[0])!= MP_ENCODE_OK) 
	{
		return false; 
	}
	else
	{
		pSm->opCMVS[counter].ucPayloadSize = 1;
		pSm->opCMVS[counter].cmvTag = EIV_CMDTAG_DMT;
		counter++;
	}

	pSm->usNumOperationalCmvs = counter;
	    
    
            
     /*
     * Init our message pending flags
     */
     pSm->mp_tx_msg_pending = FALSE;
     pSm->mp_rx_async_msg_pending = FALSE;
     pSm->mp_rx_sync_msg_pending = FALSE;
     pSm->inSM = FALSE;    
    
    eu_leaves (DBG_MSG);
   return true; 
}




/**
 * eu_write_dsp_msg - Write to DSP.
 *
 */
static void eu_write_dsp_msg (
                              eu_instance_t *ins,
                              uint16_t       addr,
                              uint16_t      *data
                             )
{
    struct urb *urb;
    devrequest *pdr;
    uint8_t *xfer_buff = NULL;

    xfer_buff = GET_KBUFFER ( ( CMV_DATA_WORDS*2 ) + sizeof(devrequest) );

    if ( !xfer_buff )
    {
        eu_err ("eu_write_dsp_msg: cannot allocate %#lx bytes\n",
                (unsigned long)(CMV_DATA_WORDS*2) + sizeof(devrequest) );
        goto byebye;
    }
    memcpy ( xfer_buff, data, CMV_DATA_WORDS*2 );
    pdr = (devrequest *) ( xfer_buff + CMV_DATA_WORDS*2 );

    /*
     * Get an URB and prepare it for submission
     */
    urb = USB_ALLOC_URB ( 0 , GFP_ATOMIC );

    FILL_USB_CTRL_REQUEST ( pdr,
                            USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                            CDC_REQ_SEND_ENCAPSULATED_CMD,
                            E4_CMD_SET_BLOCK, (addr | 0x4000), CMV_DATA_WORDS*2 ); 

    usb_fill_control_urb ( urb,
                           ins->usbdev,
                           usb_sndctrlpipe(ins->usbdev,0),
                           (uint8_t *) pdr,
                           xfer_buff,
                           CMV_DATA_WORDS * 2,
                           ctrl_urb_completion,
                           NULL );
    
    queue_ctrl_urb ( ins, urb );

  byebye:
    return ;
}


